/**********************************************************************
 Parser for STAR files
 **********************************************************************/

%{
    GlobalBlockNode *previousGlobalBlock = NULL;
    extern int line_num, col_num;   /* These filled by the scanner */
    extern int star_err_cnt;  /* starts at zero, ++'s on each error */
    extern void debug_msg( const char * ); /* contained in star.l */
    extern void finish_err_msg( StarErrID, const char * );
    extern int  last_loop_line, last_loop_col;


#ifndef STAR_SYNTAX_PRINT_DEF
#define STAR_SYNTAX_PRINT_DEF
    extern bool star_syntax_print; /* true = starlib prints syntax errors */
#endif

#ifndef STAR_WARNING_CHECK_DEF
    extern bool star_warning_check; /* false = turn off warning generation */
#define STAR_WARNING_CHECK_DEF
#endif

    extern bool err_is_printing;
#define FINISH_ERR_MSG(id, str)                               \
            if(err_is_printing)                               \
            {                                                 \
	        char tmp_str[1024] ;                          \
		sprintf( tmp_str,                             \
		         "%s (star.y:%d)", (str), __LINE__ ); \
		finish_err_msg( (id), (tmp_str) );            \
	    }

%}

// The types of the tokens and nonterminals
%union {
    TokenType         	  TOKENval;   // type of all tokens
    StarFileNode*         StarFileNodeval;
    ASTlist<BlockNode*>*     StarListNodeval;
    BlockNode *           BlockNodeval;
    GlobalBlockNode *     GlobalBlockNodeval;
    DataBlockNode*        DataBlockNodeval;
    ASTlist<DataNode*>*      DataBlockListNodeval;
    GlobalHeadingNode*    GlobalHeadingNodeval;
    DataHeadingNode*      DataHeadingNodeval;
    DataNode*             DataNodeval;
    DataItemNode*         DataItemNodeval;
    DataValueNode*        DataValueNodeval;
    DataNameNode*         DataNameNodeval;
    SaveFrameNode*        SaveFrameNodeval;
    SaveHeadingNode*      SaveHeadingNodeval;
    ASTlist<DataNode *>*     SaveFrameListNodeval;
    DataLoopNode*         DataLoopNodeval;
    ASTlist<LoopNameListNode*>* DataLoopNameListNodeval;
    ASTlist<LoopValListNode*>* DataLoopValListNodeval;
    ASTlist<DataNameNode*>*  LoopNameListNodeval;
    ASTlist<DataValueNode*>* LoopValListNodeval;
    int                      dummy;
};

//token declarations
%token <TOKENval> LOOP
%token <TOKENval> STOP
%token <TOKENval> DATA_NAME
%token <TOKENval> DATA_HEADING
%token <TOKENval> GLOBAL_HEADING
%token <TOKENval> SAVE_HEADING
%token <TOKENval> SAVE_END
%token <TOKENval> NON_QUOTED_TEXT_STRING
%token <TOKENval> DOUBLE_QUOTED_TEXT_STRING
%token <TOKENval> SINGLE_QUOTED_TEXT_STRING
%token <TOKENval> SEMICOLON_BOUNDED_TEXT_STRING 
%token <TOKENval> FRAME_CODE

//type declarations of all non-terminals
%type <StarFileNodeval>		starFile 
%type <StarListNodeval> 	starList
%type <BlockNodeval> 		block
%type <GlobalBlockNodeval>	globalBlock
%type <DataBlockNodeval> 	dataBlock
%type <DataBlockListNodeval> 	dataBlockList
%type <GlobalHeadingNodeval> 	globalHeading
%type <DataHeadingNodeval> 	dataHeading
%type <DataNodeval>		data
%type <DataItemNodeval> 	dataItem
%type <DataValueNodeval> 	dataValue
%type <SaveFrameNodeval> 	saveFrame
%type <SaveHeadingNodeval>      saveHeading
%type <SaveFrameListNodeval> 	saveFrameList
%type <DataNodeval> 		save
%type <DataLoopNodeval>   	dataLoop
%type <DataLoopNameListNodeval>  dataLoopDefList
%type <DataLoopValListNodeval>  dataLoopValList
%type <LoopNameListNodeval>      loopDefList
%type <LoopValListNodeval> 	loopValList
%type <dummy>			optionalSTOP
 

//start non-terminal declaration
%start starFile

// These shift/reduce conflicts are from the fact that the stop_ at
// the end of a loop is optional, and I am trying to do error
// recovery.  The STAR syntax is not condusive to YACC error recovery,
// due to the lack of syncronization tokens (like the semicolon in C).
// In all the cases, yacc's default behaviour of preferring the shift
// over the reduce is exactly what I want.  (It took me a while to
// verify that that is true.)
%expect 14

%%
starFile	: starList
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a starFile completed\n" );
#               endif
		    AST = new StarFileNode(new StarListNode($1));
		    AST->setLineNum( line_num );
		    AST->setColNum( col_num );
		    previousGlobalBlock = NULL ; // for the next parse.
		    last_loop_line = 0; // for the next parse.
		    last_loop_col = 0; // for the next parse.
		    
		    err_is_printing = false;
		    if( star_err_cnt > 0 )
		    {
			if( (starErrReport.end()-1) -> errID == STAR_UNKNOWN_ERR )
			    if( star_syntax_print )
				printf( "Parse error.\n" ); /* handle if last err unknown */
			// If yyparse() was able to recover from all its errors,
			// using the 'error' rules provided in this file, then
			// it returns a success code back to the caller.  That's
			// not what we want here - we want to have the yyparse()
			// return a failure code even if it recovered.  The only
			// kind of recovery we were implementing for it was just
			// enough to get it to limp along and find more than one
			// error at a time, not enough to actually produce a real
			// star tree:

			bool  has_real_non_warning_errs = false;
			for( int eIdx = 0 ; eIdx < starErrReport.size() ; eIdx++ )
			{
			    if( starErrReport[eIdx].errID < 500 )
			    {
				has_real_non_warning_errs = true;
			    }
			}
			star_err_cnt = 0; // for the next run, if there is one.
			if( has_real_non_warning_errs )
			{
			    YYABORT;
			}
		    }
		    else
		    {
			// No errors, delete the contents of the error list that
			// may have been hanging around from a previous run of
			// yyparse().
			starErrReport.erase( starErrReport.begin(), starErrReport.end() );
		    }
		    star_err_cnt = 0;
		  }
		;

starList 	: starList block                        
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a starList completed\n" );
#               endif
		    $$ = $1;
		    if( $2 != NULL )
		      $$->AddToEnd($2); 
		  }
		| /* epsilon */                         
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a starList completed (epsilon)\n" );
#               endif
		    $$ = new ASTlist<BlockNode *>;
		  }
		;

block		: globalBlock
                {
#               ifdef DEBUG
		    debug_msg( "yacc: a Block completed (version A:global)\n" );
#               endif
		    $$ = $1;
		    if( $$ != NULL )
		    {
			$$->setLineNum( $1->getLineNum() );
			$$->setColNum( $1->getColNum() ); 
		    }
                }
		| dataBlock
                {
#               ifdef DEBUG
		    debug_msg( "yacc: a Block completed (version B:datablock)\n" );
#               endif
		    $$ = $1;
		    $$->setLineNum( $1->getLineNum() );
		    $$->setColNum( $1->getColNum() ); 
		}
		| error
		{
		    $$ = NULL;
		    FINISH_ERR_MSG( EXP_GLOBAL_OR_DATA,
		                   "Expected a global_ or data_ block.");
		    yyclearin; yyerrok;
		}

		;

globalBlock	: globalHeading dataBlockList
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a globalBlock completed\n" );
#               endif
		      if( previousGlobalBlock == NULL )
		      {
			  $$ = new GlobalBlockNode($1, new DataListNode($2));
			  $$->setLineNum( $1->getLineNum() );
			  $$->setColNum( $1->getColNum() );
			  previousGlobalBlock = $$;
		      }
		      else
		      {
			  ASTlist<DataNode*> *contents = 
			      previousGlobalBlock->GiveMyDataList();
			  contents->AddToEnd( *($2) );
			  $$ = NULL;
		      }
		  }
		;

dataBlock	: dataHeading dataBlockList         
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a dataBlock completed\n" );
#               endif
		    $$ = new DataBlockNode($1, new DataListNode($2));
		    $$->setLineNum( $1->getLineNum() );
		    $$->setColNum( $1->getColNum() );
		  }
		;

dataBlockList 	: dataBlockList data             
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a DataBlockList completed\n" );
#               endif
		    $$ = $1;
		    if( $2 != NULL )
			$$->AddToEnd($2);
		  }
		| data
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a starList completed (epsilon)\n" );
#               endif
		    $$ = new ASTlist<DataNode *>;
		    if( $1 != NULL )
			$$->AddToEnd($1);
		  }
		| dataBlockList error data
		  {
#               ifdef DEBUG
		    debug_msg( "yacc: a DataBlockList completed\n" );
#               endif
		    $$ = $1;
		    if( $3 != NULL )
			$$->AddToEnd($3);
		    FINISH_ERR_MSG( EXP_DATANODE,
		                    "In a data block, encountered something "
		                    "that was not a tag/value, loop, or saveframe." );
		    yyclearin; yyerrok;
		  }
		| error data
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a starList completed (epsilon)\n" );
#               endif
		    $$ = new ASTlist<DataNode *>;
		    if( $2 != NULL )
			$$->AddToEnd($2);
		    FINISH_ERR_MSG( EXP_DATANODE,
		                    "In a data block, encountered something "
		                    "that was not a tag/value, loop, or saveframe." );
		    yyclearin; yyerrok;
		  }
		;

globalHeading	: GLOBAL_HEADING
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a globalHeading completed\n" );
#               endif
		    $$ = new GlobalHeadingNode(*($1.strVal));
		    $$->setLineNum( $1.lineNum );
		    $$->setColNum( $1.colNum );
		    delete $1.strVal;
		  }
		;

dataHeading	: DATA_HEADING
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a dataHeading completed\n" );
#               endif
		    $$ = new DataHeadingNode((*$1.strVal));
		    $$->setLineNum( $1.lineNum );
		    $$->setColNum( $1.colNum );
		    delete $1.strVal;
		  }
	        ;

data		: dataLoop 
                {
#               ifdef DEBUG
		    debug_msg( "yacc: a data completed (version A)\n" );
#               endif
		    $$ = $1;
		    $$->setLineNum( $1->getLineNum() );
		    $$->setColNum( $1->getColNum()  );
		}
		| dataItem
		{
#               ifdef DEBUG
		    debug_msg( "yacc: a data completed (version B)\n" );
#               endif
		    $$ = $1;
		    if( $1 != NULL )
		    {
			$$->setLineNum( $1->getLineNum() );
			$$->setColNum( $1->getColNum()  );
		    }
		}
		| saveFrame
		{
#               ifdef DEBUG
		    debug_msg( "yacc: a data completed (version C)\n" );
#               endif
		    $$ = $1;
		    $$->setLineNum( $1->getLineNum() );
		    $$->setColNum( $1->getColNum()  );
		}
		;

dataItem	: DATA_NAME dataValue           
		  {
#               ifdef DEBUG
		    debug_msg( "yacc: a dataItem completed\n" );
#               endif
		    DataNameNode *name = new DataNameNode(*($1.strVal));
		    name->setLineNum( $1.lineNum );
		    name->setColNum( $1.colNum );

		    $$ = new DataItemNode(name,$2);
		    $$->setLineNum( name->getLineNum() );
		    $$->setColNum( name->getColNum() );
		    delete $1.strVal;
		  }
	        ;

dataValue	: NON_QUOTED_TEXT_STRING
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a dataValue completed (version A)\n" );
#               endif
		      $$ = new DataValueNode(  (*$1.strVal),
			                       DataValueNode::NON );
		      $$->setLineNum( $1.lineNum );
		      $$->setColNum( $1.colNum );
		      delete $1.strVal;
		  }
		| DOUBLE_QUOTED_TEXT_STRING
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a dataValue completed (version B)\n" );
#               endif
		      $$ = new DataValueNode(  *($1.strVal),
			                       DataValueNode::DOUBLE );
		      $$->setLineNum( $1.lineNum );
		      $$->setColNum( $1.colNum );
		      delete $1.strVal;
		  }
		| SINGLE_QUOTED_TEXT_STRING
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a dataValue completed (version C)\n" );
#               endif
		      $$ = new DataValueNode(  *($1.strVal),
			                       DataValueNode::SINGLE );
		      $$->setLineNum( $1.lineNum );
		      $$->setColNum( $1.colNum );
		      delete $1.strVal;
		  }
		| SEMICOLON_BOUNDED_TEXT_STRING
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a dataValue completed (version D)\n" );
#               endif
		      char *str = strdup( (*($1.strVal)).c_str() );

		      $$ = new DataValueNode(  str, DataValueNode::SEMICOLON );
		      $$->setLineNum( $1.lineNum );
		      $$->setColNum( $1.colNum );
		      // Check for warning: keywords inside the semicolon
		      // string:
		      char *keyword_pos = NULL;
		      char keyword[20];
		      keyword[0] = '\0';
		      char errMsg[200];
		      if( (keyword_pos = strstr( str, "stop_" )) != NULL )
			  strcpy( keyword, "stop_" );
		      else if( (keyword_pos = strstr( str, "loop_" )) != NULL )
			  strcpy( keyword, "loop_" );
		      else if( (keyword_pos = strstr( str, "save_" )) != NULL )
			  strcpy( keyword, "save_" );
		      else if( (keyword_pos = strstr( str, "data_" )) != NULL )
			  strcpy( keyword, "data_" );
		      if( keyword_pos != NULL )
		      {
			  int   start_line=0, keyword_line=0, cur_line=0;
			  char  *ch=NULL;

			  // Calculate the line numbers to report in the
			  // error message (counting backwards):
			  for(   cur_line = $1.lineNum,
			         ch = &( str[strlen(str)-1] )
			         ;
			         ch >= &(str[0])
				 ;
				 ch-- )
			  {
			      if( *ch == '\n' )
				  cur_line--;
			      if( ch == keyword_pos )
				  keyword_line = cur_line;
			  }
			  start_line = cur_line;

			  if( star_warning_check )
			  {
			      sprintf( errMsg, 
				      "Possible runaway value from line %d:\n"
				      "\t(This suspicion is due to the '%s' "
				      "on line %d.)",
				      start_line, keyword, keyword_line );
			      yyerror( "WARNING" );
			      FINISH_ERR_MSG( WARN_KEYWORD_IN_VALUE, errMsg);
			  }
		      }
		      free(str);
		      delete $1.strVal;
		  }
		| FRAME_CODE
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a dataValue completed (version E)\n" );
#               endif
		      $$ = new DataValueNode(  *($1.strVal),
			                       DataValueNode::FRAMECODE );
		      $$->setLineNum( $1.lineNum );
		      $$->setColNum( $1.colNum );
		      delete $1.strVal;
		  }
		;

saveFrame	: saveHeading saveFrameList SAVE_END
		  {
#               ifdef DEBUG
		    debug_msg( "yacc: a saveFrame completed\n" );
#               endif
		    $$ = new SaveFrameNode($1, new SaveFrameListNode($2));
		    $$->setLineNum( $1->getLineNum() );
		    $$->setColNum( $1->getColNum() );
		  }
		| saveHeading SAVE_END /*empty contents*/
		  { 
#               ifdef DEBUG
		    debug_msg( "yacc: a saveFrame completed\n" );
#               endif
		    $$ = new SaveFrameNode($1->myName() );
		    $$->setLineNum( $1->getLineNum() );
		    $$->setColNum( $1->getColNum() );
		  }
		;

saveHeading     : SAVE_HEADING
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a saveHeading completed\n" );
#               endif
		    $$ = new SaveHeadingNode( *($1.strVal ));
		    $$->setLineNum( $1.lineNum );
		    $$->setColNum( $1.colNum );
		    delete $1.strVal;
		  }

saveFrameList 	: saveFrameList save           
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a saveFrameList completed\n" );
#               endif
		      $$ = $1;
		      if( $2 != NULL )
			  $$->AddToEnd($2);
		  }
		| saveFrameList error save
		  { 
#               ifdef DEBUG
		    debug_msg( "yacc: a saveFrameList completed (err 1)\n" );
#               endif
		      $$ = $1;
		      if( $3 != NULL )
			  $$->AddToEnd($3);
		    FINISH_ERR_MSG( EXP_ITEM_OR_LOOP,
		                    "Encountered something that was not a loop_ or tag/value.");
		    yyclearin; yyerrok;
		  }
		| error save
		  { 
#               ifdef DEBUG
		    debug_msg( "yacc: a saveFrameList completed (err 2)\n" );
#               endif
		    $$ == new ASTlist<DataNode *>;
		    if( $2 != NULL )
			$$->AddToEnd($2);
		    FINISH_ERR_MSG( EXP_ITEM_OR_LOOP,
		                    "Encountered something that was not a loop_ or tag/value.");
		    yyclearin; yyerrok;
		  }
		| save
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a saveFrameList completed (epsilon)\n" );
#               endif
		     $$ = new ASTlist<DataNode *>;
		     if( $1 != NULL )
			 $$->AddToEnd($1);
		  }
		;

save		: dataLoop
                {
#               ifdef DEBUG
		    debug_msg( "yacc: a save completed (version A)\n" );
#               endif
		    $$ = $1;
		    $$->setLineNum( $1->getLineNum() );
		    $$->setColNum( $1->getColNum() );
		}
		| dataItem
		{
#               ifdef DEBUG
		    debug_msg( "yacc: a save completed (version B)\n" );
#               endif
		    $$ = $1;
		    if( $1 != NULL )
		    {
			$$->setLineNum( $1->getLineNum() );
			$$->setColNum( $1->getColNum() );
		    }
		}
		;

dataLoop	: LOOP dataLoopDefList dataLoopValList
                  { 
#               ifdef DEBUG
		    debug_msg( "yacc: a dataLoop completed\n" );
#               endif
		    DataLoopNameListNode  *defL = new DataLoopDefListNode($2);
		    DataLoopValListNode  *valL = new DataLoopValListNode($3);
#               ifdef DEBUG
		    defL->debugPrint();
		    valL->debugPrint();
#               endif
		    $2->Reset();
		    if( $2->Length() > 0 )
		    {
			defL->setLineNum( $2->Current()->getLineNum() );
			defL->setColNum( $2->Current()->getColNum() );
		    }
		    $3->Reset();
		    if( $3->Length() > 0 )
		    {
			valL->setLineNum( $3->Current()->getLineNum() );
			valL->setColNum( $3->Current()->getColNum() );
		    }
		    $$ = new DataLoopNode(defL, valL);
		    $$->setLineNum( $1.lineNum );
		    $$->setColNum( $1.colNum );
		    last_loop_line = line_num;
		    last_loop_col = col_num;

		    // change the tab flag depending on if the first
		    // row is linear or tabulated:
		    // ---------------------------------------------

		    $$->getVals().setTabFlag( false ); // the default

		    if( $$->getVals().size() > 0 )
		    {
			int i;
		        int prev_line = -9999;
			for( i = 0 ; i < ($$->getVals()[0])->size() ; i++ )
			{
			    if( prev_line ==
			           (*($$->getVals()[0]))[i]->getLineNum() )
			    {
				$$->setTabFlag( true );
				break;
			    }
			    prev_line = (*($$->getVals()[0]))[i]->getLineNum();
			}
		    }

		    char errMsg[256];
		    List<int> *badAligns =
			$$->getValsPtr()->checkValueAlignment();
		    for( int aIdx = 0 ; aIdx < badAligns->size() ; aIdx++ )
		    {
			if( star_warning_check )
			{
			    yyerror( "WARNING" );
			    sprintf( errMsg, "In the previous loop, misaligned "
				     "values for the row starting on line %d.",
				     (*badAligns)[aIdx] );
			    FINISH_ERR_MSG( WARN_LOOP_VALUES_MISALIGNED, errMsg );
			}
		    }
		    delete badAligns;
		  }
		;

dataLoopDefList : dataLoopDefList LOOP loopDefList  
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a dataLoopDefList completed (version A)\n" );
#               endif
		    $$ = $1;
		    LoopNameListNode *ldlN = new LoopDefListNode($3);
		    $3->Reset();
		    if( $3->Length() > 0 )
		    {
		        ldlN->setLineNum( $3->Current()->getLineNum() );
		        ldlN->setColNum( $3->Current()->getColNum() );
		    }
		    $$->AddToEnd( ldlN );
		  }
                | loopDefList   
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a dataLoopDefList completed (version B)\n" );
#               endif
		    $$ = new ASTlist<LoopNameListNode *>;
		    LoopNameListNode *ldlN = new LoopDefListNode($1);
		    $1->Reset();
		    if( $1->Length() > 0 )
		    {
		        ldlN->setLineNum( $1->Current()->getLineNum() );
		        ldlN->setColNum( $1->Current()->getColNum() );
		    }
		    $$->AddToEnd( ldlN );
		  }
                ;

dataLoopValList : dataLoopValList loopValList optionalSTOP   
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a dataLoopValList completed (version A)\n" );
#               endif
		    $$ = $1;
		    LoopValListNode *lvlN = new LoopValListNode($2);
		    $2->Reset();
		    if( $2->Length() > 0 )
		    {
			lvlN->setLineNum( $2->Current()->getLineNum() );
			lvlN->setColNum( $2->Current()->getColNum() );
		    }
		    $$->AddToEnd( lvlN );
		  }
                | loopValList optionalSTOP                
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a dataLoopValList completed (version B)\n" );
#               endif
		    $$ = new ASTlist<LoopValListNode *>;
		    LoopValListNode *lvlN = new LoopValListNode($1);
		    $1->Reset();
		    if( $1->Length() > 0 )
		    {
			lvlN->setLineNum( $1->Current()->getLineNum() );
			lvlN->setColNum( $1->Current()->getColNum() );
		    }
		    $$->AddToEnd( lvlN );
		  }
                | dataLoopValList STOP           
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a dataLoopValList completed (version C)\n" );
#               endif
		    $$ = $1;
		    LoopValListNode *emptyStart = new LoopValListNode;
		    $$->AddToEnd(emptyStart);
		  }
                ;

loopDefList	: loopDefList DATA_NAME
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a loopDefList completed (version A)\n" );
#               endif
		      DataNameNode *name = new DataNameNode( *($2.strVal) );
		      name->setLineNum( $2.lineNum );
		      name->setColNum( $2.colNum );

		      $$ = $1;
		      $$->AddToEnd( name );
		      delete $2.strVal;
		  }
		| DATA_NAME
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a loopDefList completed (version B)\n" );
#               endif
		    DataNameNode *name = new DataNameNode( *($1.strVal) );
		    name->setLineNum( $1.lineNum );
		    name->setColNum( $1.colNum );

		    $$ = new ASTlist<DataNameNode *>;
		    $$->AddToEnd( name ); 
		    delete $1.strVal;
		  }
		;

optionalSTOP	: /*nothing or STOP*/
		  {
#               ifdef DEBUG
		    debug_msg( "yacc: a optionalSTOP completed (epsilon)\n" );
#               endif
		      $$ = 1;
		  }
                |  STOP
		  {
#               ifdef DEBUG
		    debug_msg( "yacc: a optionalSTOP completed (version B)\n" );
#               endif
		      $$ = 1;
		  }
		;

loopValList     : loopValList dataValue     
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a loopValList completed (version A)\n" );
#               endif
		      $$ = $1; $$->AddToEnd($2);
		  }
		| dataValue                 
                  {
#               ifdef DEBUG
		    debug_msg( "yacc: a loopValList completed (version B)\n" );
#               endif
		      $$ = new ASTlist<DataValueNode *>;

		      $$->AddToEnd($1); 
		  }
		;
